#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


# Run manually with python2 and python3 src/python/test/test_extract_assay.py

import unittest
import dxpy
import os
import subprocess
import json
import sys

from unittest.mock import patch
from io import StringIO

from parameterized import parameterized

from dxpy_testutil import cd
from dxpy.dx_extract_utils.filter_to_payload import (
    retrieve_geno_bins,
    basic_filter,
    location_filter,
    generate_assay_filter,
    final_payload,
    validate_JSON,
)
from dxpy.dx_extract_utils.germline_utils import (
    filter_results,
    _produce_loci_dict,
    infer_genotype_type
)
from dxpy.cli.dataset_utilities import (
    DXDataset,
    resolve_validate_record_path,
    get_assay_name_info,
)
from dxpy.dx_extract_utils.input_validation import validate_filter_applicable_genotype_types


python_version = sys.version_info.major

dirname = os.path.dirname(__file__)


class TestDXExtractAssay(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        test_project_name = "dx-toolkit_test_data"
        cls.test_v1_record = "{}:/Extract_Assay_Germline/test01_dataset".format(
            test_project_name
        )
        cls.test_record = "{}:/Extract_Assay_Germline/test01_v1_0_1_dataset".format(
            test_project_name
        )
        cls.test_non_alt_record = "{}:/Extract_Assay_Germline/test03_dataset".format(
            test_project_name
        )
        cls.output_folder = os.path.join(dirname, "extract_assay_germline/test_output/")
        cls.malformed_json_dir = os.path.join(dirname, "ea_malformed_json")
        cls.proj_id = list(
            dxpy.find_projects(describe=False, level="VIEW", name=test_project_name)
        )[0]["id"]
        cd(cls.proj_id + ":/")

    ############
    # Unit Tests
    ############

    # Test retrieve_geno_bins
    def test_get_assay_name_info(self):
        # Set to true for the list assay utilities response instead of the normal functionality
        list_assays = False
        # When assay name is none, function looks for and selects first assay of type somatic that it finds
        assay_name = None
        friendly_assay_type = "germline"
        project, entity_result, resp, dataset_project = resolve_validate_record_path(
            self.test_record
        )
        dataset_id = resp["dataset"]
        rec_descriptor = DXDataset(dataset_id, project=dataset_project).get_descriptor()
        # Expected Results
        expected_assay_name = "test01_assay"
        expected_assay_id = "cc5dcc31-000c-4a2c-b225-ecad6233a0a3"
        expected_ref_genome = "GRCh38.92"
        expected_additional_descriptor_info = {
            "exclude_refdata": True,
            "exclude_halfref": True,
            "exclude_nocall": True,
            "genotype_type_table": "genotype_alt_read_optimized",
        }

        (
            selected_assay_name,
            selected_assay_id,
            selected_ref_genome,
            additional_descriptor_info,
        ) = get_assay_name_info(
            list_assays=False,
            assay_name=assay_name,
            path=self.test_record,
            friendly_assay_type=friendly_assay_type,
            rec_descriptor=rec_descriptor,
        )

        self.assertEqual(expected_assay_name, selected_assay_name)
        self.assertEqual(expected_assay_id, selected_assay_id)
        self.assertEqual(expected_ref_genome, selected_ref_genome)
        self.assertEqual(expected_additional_descriptor_info, additional_descriptor_info)

    def test_retrieve_geno_bins(self):
        # list_of_genes, project, genome_reference
        list_of_genes = ["ENSG00000173213"]
        genome_reference = "GRCh38.92"
        expected_output = [{"chr": "18", "start": "47390", "end": "49557"}]
        self.assertEqual(
            retrieve_geno_bins(list_of_genes, self.proj_id, genome_reference),
            expected_output,
        )

    def test_basic_filter_allele(self):
        table = "allele"
        friendly_name = "rsid"
        values = ["rs1342568097"]
        genome_reference = "GRCh38.92"

        expected_output = {
            "allele$dbsnp151_rsid": [{"condition": "any", "values": ["rs1342568097"]}]
        }

        self.assertEqual(
            basic_filter(table, friendly_name, values, self.proj_id, genome_reference),
            expected_output,
        )

    def test_basic_filter_annotation(self):
        table = "annotation"
        friendly_name = "gene_id"
        values = ["ENSG00000173213"]
        genome_reference = "GRCh38.92"

        expected_output = {
            "annotation$gene_id": [
                {
                    "condition": "in",
                    "values": ["ENSG00000173213"],
                    "geno_bins": [{"chr": "18", "start": "47390", "end": "49557"}],
                }
            ]
        }

        self.assertEqual(
            basic_filter(table, friendly_name, values, self.proj_id, genome_reference),
            expected_output,
        )

    def test_basic_filter_genotype(self):
        table = "genotype"
        friendly_name = "allele_id"
        values = ["18_47361_T_G"]
        genome_reference = "GRCh38.92"

        expected_output = {
            "allele$a_id": [{"condition": "in", "values": ["18_47361_T_G"]}]
        }

        self.assertEqual(
            basic_filter(table, friendly_name, values, self.proj_id, genome_reference),
            expected_output,
        )

    def test_location_filter(self):
        location_list = [
            {
                "chromosome": "18",
                "starting_position": "47361",
                "ending_position": "47364",
            }
        ]

        expected_output = {
            "allele$a_id": [
                {
                    "condition": "in",
                    "values": [],
                    "geno_bins": [{"chr": "18", "start": 47361, "end": 47364}],
                }
            ]
        }

        self.assertEqual(location_filter(location_list, "allele"), expected_output)

    def test_genotype_location_filter(self):
        location_list = [
            {
                "chromosome": "18",
                "starting_position": "47361",
            }
        ]

        expected_output = {
            "genotype$a_id": [
                {
                    "condition": "in",
                    "values": [],
                    "geno_bins": [{"chr": "18", "start": 47361, "end": 47361}],
                }
            ]
        }

        self.assertEqual(location_filter(location_list, "genotype"), expected_output)

    def test_generate_assay_filter(self):
        # A small payload, uses allele_rsid.json
        full_input_dict = {"rsid": ["rs1342568097"]}
        name = "test01_assay"
        id = "c6e9c0ea-5752-4299-8de2-8620afba7b82"
        genome_reference = "GRCh38.92"
        filter_type = "allele"

        expected_output = {
            "assay_filters": {
                "name": "test01_assay",
                "id": "c6e9c0ea-5752-4299-8de2-8620afba7b82",
                "filters": {
                    "allele$dbsnp151_rsid": [
                        {"condition": "any", "values": ["rs1342568097"]}
                    ]
                },
                "logic": "and",
            }
        }

        self.assertEqual(
            generate_assay_filter(
                full_input_dict,
                name,
                id,
                self.proj_id,
                genome_reference,
                filter_type,
            ),
            expected_output,
        )

    def test_final_payload(self):
        full_input_dict = {"rsid": ["rs1342568097"]}
        name = "test01_assay"
        id = "c6e9c0ea-5752-4299-8de2-8620afba7b82"
        genome_reference = "GRCh38.92"
        filter_type = "allele"

        expected_output_payload = {
            "project_context": self.proj_id,
            "order_by": [{"allele_id":"asc"}],
            "fields": [
                {"allele_id": "allele$a_id"},
                {"chromosome": "allele$chr"},
                {"starting_position": "allele$pos"},
                {"ref": "allele$ref"},
                {"alt": "allele$alt"},
                {"rsid": "allele$dbsnp151_rsid"},
                {"allele_type": "allele$allele_type"},
                {"dataset_alt_freq": "allele$alt_freq"},
                {"gnomad_alt_freq": "allele$gnomad201_alt_freq"},
                {"worst_effect": "allele$worst_effect"},
            ],
            "adjust_geno_bins": False,
            "raw_filters": {
                "assay_filters": {
                    "name": "test01_assay",
                    "id": "c6e9c0ea-5752-4299-8de2-8620afba7b82",
                    "filters": {
                        "allele$dbsnp151_rsid": [
                            {"condition": "any", "values": ["rs1342568097"]}
                        ]
                    },
                    "logic": "and",
                }
            },
            "is_cohort": True,
            "distinct": True,
        }

        expected_output_fields = [
            "allele_id",
            "chromosome",
            "starting_position",
            "ref",
            "alt",
            "rsid",
            "allele_type",
            "dataset_alt_freq",
            "gnomad_alt_freq",
            "worst_effect",
        ]

        test_payload, test_fields = final_payload(
            full_input_dict,
            name,
            id,
            self.proj_id,
            genome_reference,
            filter_type,
        )

        self.assertEqual(
            test_payload,
            expected_output_payload,
        )
        self.assertEqual(test_fields, expected_output_fields)

    def test_validate_json(self):
        filter = {
            "rsid": ["rs1342568097"],
            "type": ["SNP", "Del", "Ins"],
            "dataset_alt_af": {"min": 1e-05, "max": 0.5},
            "gnomad_alt_af": {"min": 1e-05, "max": 0.5},
        }
        type = "allele"

        try:
            validate_JSON(filter, type)
        except:
            self.fail("This just needs to complete without error")

    def test_malformed_json(self):
        for filter_type in ["allele", "annotation", "genotype"]:
            malformed_json_filenames = os.listdir(
                os.path.join(self.malformed_json_dir, filter_type)
            )
            for name in malformed_json_filenames:
                file_path = os.path.join(self.malformed_json_dir, filter_type, name)
                with open(file_path, "r") as infile:
                    filter = json.load(infile)
                with self.assertRaises(SystemExit) as cm:
                    validate_JSON(filter, filter_type)
                self.assertEqual(cm.exception.code, 1)

    def test_bad_rsid(self):
        filter = {"rsid": ["rs1342568097", "rs1342568098"]}
        test_project = "dx-toolkit_test_data"
        test_record = f"{test_project}:Extract_Assay_Germline/test01_v1_0_1_dataset"

        command = [
            "dx",
            "extract_assay",
            "germline",
            test_record,
            "--retrieve-allele",
            json.dumps(filter),
            "-o",
            "-",
        ]
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, universal_newlines=True
        )
        result = process.communicate()[0]
        expected_result = (
            '18_47359_G_T\t18\t47359\tG\tT\t["rs1342568097"]\tSNP\t0.2\t2.66771e-05'
        )
        self.assertIn(expected_result, result)

    def test_duplicate_rsid(self):
        table = "allele"
        friendly_name = "rsid"
        values = ["rs1342568097", "rs1342568097"]
        genome_reference = "GRCh38.92"

        expected_output = {
            "allele$dbsnp151_rsid": [{"condition": "any", "values": ["rs1342568097"]}]
        }

        self.assertEqual(
            basic_filter(table, friendly_name, values, self.proj_id, genome_reference),
            expected_output,
        )

    # Test filters for exclusion options
    def test_no_call_warning(self):
        # This is a test of the validate_filter_applicable_genotype_types function "no-call", "ref" requested
        filter_dict = {"genotype_type": ["no-call", "ref"]}
        exclude_nocall = True
        exclude_refdata = True
        infer_nocall = False
        expected_warnings = [
            "WARNING: Filter requested genotype type 'no-call', genotype entries of this type were not ingested in the provided dataset and the --infer-nocall flag is not set!",
            "WARNING: Filter requested genotype type 'ref', genotype entries of this type were not ingested in the provided dataset and the --infer-ref flag is not set!"
            ]
        with patch("sys.stderr", new=StringIO()) as fake_err:
            validate_filter_applicable_genotype_types(
                infer_nocall, infer_ref=False, filter_dict=filter_dict,
                exclude_refdata=exclude_refdata, exclude_nocall=exclude_nocall, exclude_halfref=False
            )
            output = fake_err.getvalue().strip()
            for warning in expected_warnings:
                self.assertIn(warning, output)

    def test_no_genotype_type_warning(self):
        # This is a test of the validate_filter_applicable_genotype_types function no genotype type requested
        filter_dict = {"genotype_type": []}
        exclude_nocall = True

        with patch("sys.stderr", new=StringIO()) as fake_err:
            validate_filter_applicable_genotype_types(
                infer_nocall=False, infer_ref=False, filter_dict=filter_dict,
                exclude_refdata=False, exclude_nocall=exclude_nocall, exclude_halfref=False
            )
            output = fake_err.getvalue().strip()
            self.assertEqual(output, "WARNING: No genotype type requested in the filter. All genotype types will be returned. Genotype entries of type 'no-call' were not ingested in the provided dataset and the --infer-nocall flag is not set!")

    def test_no_genotype_type_warning_exclude_halfref(self):
        # This is a test of the validate_filter_applicable_genotype_types function half genotype type requested
        filter_dict = {"genotype_type": []}
        exclude_halfref = True

        with patch("sys.stderr", new=StringIO()) as fake_err:
            validate_filter_applicable_genotype_types(
                infer_nocall=False, infer_ref=False, filter_dict=filter_dict,
                exclude_refdata=False, exclude_nocall=False, exclude_halfref=exclude_halfref
            )
            output = fake_err.getvalue().strip()
            self.assertEqual(output, "WARNING: No genotype type requested in the filter. All genotype types will be returned.  'half-ref' genotype entries (0/.) were not ingested in the provided dataset!")

    def test_filter_results(self):
        # Define sample input data
        results = [
            {
                "sample_id": "SAMPLE_1",
                "allele_id": "1_1076145_A_AT",
                "locus_id": "1_1076145_A_T",
                "chromosome": "1",
                "starting_position": 1076145,
                "ref": "A",
                "alt": "AT",
                "genotype_type": "het-alt",
            },
            {
                "sample_id": "SAMPLE_2",
                "allele_id": "1_1076146_A_AT",
                "locus_id": "1_1076146_A_T",
                "chromosome": "1",
                "starting_position": 1076146,
                "ref": "A",
                "alt": "AT",
                "genotype_type": "het-alt",
            },
            {
                "sample_id": "SAMPLE_3",
                "allele_id": "1_1076147_A_AT",
                "locus_id": "1_1076147_A_T",
                "chromosome": "1",
                "starting_position": 1076147,
                "ref": "A",
                "alt": "AT",
                "genotype_type": "ref",
            },
        ]
        # Call the function to filter the results
        filtered_results = filter_results(
            results=results, key="genotype_type", restricted_values=["het-alt"]
        )

        # Define the expected output
        expected_output = [
            {
                "sample_id": "SAMPLE_3",
                "allele_id": "1_1076147_A_AT",
                "locus_id": "1_1076147_A_T",
                "chromosome": "1",
                "starting_position": 1076147,
                "ref": "A",
                "alt": "AT",
                "genotype_type": "ref",
            },
        ]

        # Assert that the filtered results match the expected output
        self.assertEqual(filtered_results, expected_output)

    def test_produce_loci_dict(self):
        # Define the input data
        loci = [
            {
                "locus_id": "18_47361_A_T",
                "chromosome": "18",
                "starting_position": 47361,
                "ref": "A",
            },
            {
                "locus_id": "X_1000_C_A",
                "chromosome": "X",
                "starting_position": 1000,
                "ref": "C",
            },
            {
                "locus_id": "1_123_A_.",
                "chromosome": "1",
                "starting_position": 123,
                "ref": "A",
            },
        ]
        results_entries = [
            {
                "locus_id": "18_47361_A_T",
                "allele_id": "18_47361_A_T",
                "sample_id": "sample1",
                "chromosome": "18",
                "starting_position": 47361,
                "ref": "A",
                "alt": "T",
            },
            {
                "locus_id": "18_47361_A_T",
                "allele_id": "18_47361_A_G",
                "sample_id": "sample2",
                "chromosome": "18",
                "starting_position": 47361,
                "ref": "A",
                "alt": "G",
            },
            {
                "locus_id": "X_1000_C_A",
                "allele_id": "X_1000_C_A",
                "sample_id": "sample1",
                "chromosome": "X",
                "starting_position": 1000,
                "ref": "C",
                "alt": "A",
            },
        ]

        # Define the expected output
        expected_output = {
            "18_47361_A_T": {
                "samples": {"sample1", "sample2"},
                "entry": {
                    "allele_id": None,
                    "locus_id": "18_47361_A_T",
                    "chromosome": "18",
                    "starting_position": 47361,
                    "ref": "A",
                    "alt": None,
                },
            },
            "X_1000_C_A": {
                "samples": {"sample1"},
                "entry": {
                    "allele_id": None,
                    "locus_id": "X_1000_C_A",
                    "chromosome": "X",
                    "starting_position": 1000,
                    "ref": "C",
                    "alt": None,
                },
            },
            "1_123_A_.": {
                "samples": set(),
                "entry": {
                    "allele_id": None,
                    "locus_id": "1_123_A_.",
                    "chromosome": "1",
                    "starting_position": 123,
                    "ref": "A",
                    "alt": None,
                },
            },
        }

        # Call the function
        result = _produce_loci_dict(loci, results_entries)

        # Assert the result
        self.assertEqual(result, expected_output)

    def test_infer_genotype_type(self):
        samples = ["SAMPLE_1", "SAMPLE_2", "SAMPLE_3"]
        loci = [
            {
                "locus_id": "1_1076145_A_T",
                "chromosome": "1",
                "starting_position": 1076145,
                "ref": "A",
            },
            {
                "locus_id": "2_1042_G_CC",
                "chromosome": "2",
                "starting_position": 1042,
                "ref": "G",
            },
        ]
        result_entries = [
            {
                "sample_id": "SAMPLE_2",
                "allele_id": "1_1076145_A_AT",
                "locus_id": "1_1076145_A_T",
                "chromosome": "1",
                "starting_position": 1076145,
                "ref": "A",
                "alt": "AT",
                "genotype_type": "het-alt",
            },
            {
                "sample_id": "SAMPLE_3",
                "allele_id": "1_1076145_A_T",
                "locus_id": "1_1076145_A_T",
                "chromosome": "1",
                "starting_position": 1076145,
                "ref": "A",
                "alt": "T",
                "genotype_type": "hom-ref",
            },
        ]
        type_to_infer = "no-call"

        expected_output = [
            {
                "sample_id": "SAMPLE_1",
                "allele_id": None,
                "locus_id": "1_1076145_A_T",
                "chromosome": "1",
                "starting_position": 1076145,
                "ref": "A",
                "alt": None,
                "genotype_type": "no-call",
            },
            {
                "sample_id": "SAMPLE_1",
                "allele_id": None,
                "locus_id": "2_1042_G_CC",
                "chromosome": "2",
                "starting_position": 1042,
                "ref": "G",
                "alt": None,
                "genotype_type": "no-call",
            },
            {
                "sample_id": "SAMPLE_2",
                "allele_id": None,
                "locus_id": "2_1042_G_CC",
                "chromosome": "2",
                "starting_position": 1042,
                "ref": "G",
                "alt": None,
                "genotype_type": "no-call",
            },
            {
                "sample_id": "SAMPLE_3",
                "allele_id": None,
                "locus_id": "2_1042_G_CC",
                "chromosome": "2",
                "starting_position": 1042,
                "ref": "G",
                "alt": None,
                "genotype_type": "no-call",
            },
        ]

        output = infer_genotype_type(samples, loci, result_entries, type_to_infer)
        self.assertEqual(output, result_entries + expected_output)
    ##########
    # Normal Command Lines
    ##########

    def test_json_help(self):
        """Check successful call of help for the retrieve allele filter"""
        command = ["dx", "extract_assay", "germline", "fakepath", "--retrieve-allele", "--json-help"]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        expected_help = """# Filters and respective definitions
#
#  rsid: rsID associated with an allele or set of alleles. If multiple values
#  are provided, the conditional search will be, "OR." For example, ["rs1111",
#  "rs2222"], will search for alleles which match either "rs1111" or "rs2222".
#  String match is case sensitive. Duplicate values are permitted and will be
#  handled silently.
#
#  type: Type of allele. Accepted values are "SNP", "Ins", "Del", "Mixed". If
#  multiple values are provided, the conditional search will be, "OR." For
#  example, ["SNP", "Ins"], will search for variants which match either "SNP"
#  or "Ins". String match is case sensitive.
#
#  dataset_alt_af: Dataset alternate allele frequency, a json object with
#  empty content or two sets of key/value pair: {min: 0.1, max:0.5}. Accepted
#  numeric value for each key is between and including 0 and 1.  If a user
#  does not want to apply this filter but still wants this information in the
#  output, an empty json object should be provided.
#
#  gnomad_alt_af: gnomAD alternate allele frequency. a json object with empty
#  content or two sets of key/value pair: {min: 0.1, max:0.5}. Accepted value
#  for each key is between 0 and 1. If a user does not want to apply this
#  filter but still wants this information in the output, an empty json object
#  should be provided.
#
#  location: Genomic range in the reference genome where the starting position
#  of alleles fall into. If multiple values are provided in the list, the
#  conditional search will be, "OR." String match is case sensitive.
#
# JSON filter template for --retrieve-allele
{
  "rsid": ["rs11111", "rs22222"],
  "type": ["SNP", "Del", "Ins"],
  "dataset_alt_af": {"min": 0.001, "max": 0.05},
  "gnomad_alt_af": {"min": 0.001, "max": 0.05},
  "location": [
    {
      "chromosome": "1",
      "starting_position": "10000",
      "ending_position": "20000"
    },
    {
      "chromosome": "X",
      "starting_position": "500",
      "ending_position": "1700"
    }
  ]
}
"""
        self.assertEqual(expected_help, process.communicate()[0])

    def test_generic_help(self):
        """Test the generic help message"""
        command = "dx extract_assay germline -h > /dev/null"
        subprocess.check_call(command, shell=True)

    def test_list_assays(self):
        command = ["dx", "extract_assay", "germline", self.test_record, "--list-assays"]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        self.assertEqual("test01_assay", process.communicate()[0].strip())

    def test_assay_name(self):
        """A test of the --assay-name functionality, returns the same output."""
        allele_rsid_filter = json.dumps({"rsid": ["rs1342568097"]})
        command1 = ["dx", "extract_assay", "germline", self.test_record, "--assay-name", "test01_assay", "--retrieve-allele", allele_rsid_filter, "-o", "-"]
        process1 = subprocess.Popen(command1, stdout=subprocess.PIPE, universal_newlines=True)
        command2 = ["dx", "extract_assay", "germline", self.test_record, "--retrieve-allele", allele_rsid_filter, "-o", "-"]
        process2 = subprocess.Popen(command2, stdout=subprocess.PIPE, universal_newlines=True)
        self.assertEqual(process1.communicate(), process2.communicate())

    @parameterized.expand([
        ("test_record", ["ref", "het-ref", "hom", "het-alt", "half", "no-call"]),
        ("test_v1_record", ["het-ref", "hom", "het-alt", "half"]),
    ])
    def test_retrieve_genotype(self, record, genotype_types):
        """Testing --retrieve-genotype functionality"""
        allele_genotype_type_filter = json.dumps({
            "allele_id": ["18_47408_G_A"], 
            "genotype_type": genotype_types,
            })
        expected_result = "sample_1_3\t18_47408_G_A\t18_47408_G_A\t18\t47408\tG\tA\thet-ref"
        command = ["dx", "extract_assay", "germline", getattr(self, record), "--retrieve-genotype", allele_genotype_type_filter, "-o", "-"]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        self.assertIn(expected_result, process.communicate()[0])

    @unittest.skip("Vizserver implementation of PMUX-1652 needs to be deployed")
    def test_retrieve_non_alt_genotype(self):
        """Testing --retrieve-genotype functionality"""
        location_genotype_type_filter = json.dumps({
            "location": [{
                "chromosome": "20",
                "starting_position": "14370",
            }],
            "genotype_type": ["ref", "half", "no-call"]
        })
        # not a comprehensive list
        expected_results = [
            "S01_m_m\t\t20_14370_G_A\t20\t14370\tG\t\tno-call",
            "S02_m_0\t\t20_14370_G_A\t20\t14370\tG\t\thalf",
            "S06_0_m\t\t20_14370_G_A\t20\t14370\tG\t\thalf",
            "S07_0_0\t\t20_14370_G_A\t20\t14370\tG\t\tref",
        ]
        command = ["dx", "extract_assay", "germline", self.test_non_alt_record, "--retrieve-genotype", location_genotype_type_filter, "-o", "-"]
        process = subprocess.Popen(command, stdout=subprocess.PIPE, universal_newlines=True)
        result = process.communicate()[0]
        [self.assertIn(expected_result, result) for expected_result in expected_results]

    ###########
    # Malformed command lines
    ###########

    def test_filter_mutex(self):
        print("testing filter mutex")
        """Ensure that the failure mode of multiple filter types being provided is caught"""
        # Grab two random filter JSONs of different types
        allele_json = '{"rsid": ["rs1342568097"]}'
        annotation_json = '{"allele_id": ["18_47408_G_A"]}'
        command = ["dx", "extract_assay", "germline", self.test_record,  "--retrieve-allele", allele_json, "--retrieve-annotation", annotation_json, "-o", "-"]

        process = subprocess.Popen(command, stderr=subprocess.PIPE, universal_newlines=True)
        expected_error_message = "dx extract_assay germline: error: argument --retrieve-annotation: not allowed with argument --retrieve-allele"
        self.assertTrue(expected_error_message in process.communicate()[1])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import os
import dxpy
import json
import argparse
from glob import glob
import shutil
import tempfile
from functools import partial

from dxpy.nextflow.nextflow_templates import (get_nextflow_dxapp, get_nextflow_src)
from dxpy.nextflow.nextflow_utils import (get_template_dir, write_exec, write_dxapp, get_importer_name,
                                          create_readme, get_nested, get_allowed_extra_fields_mapping)
from dxpy.cli.exec_io import parse_obj
from dxpy.cli import try_call
from dxpy.utils.resolver import resolve_existing_path


parser = argparse.ArgumentParser(description="Uploads a DNAnexus App.")


def build_pipeline_with_npi(
        repository=None,
        tag=None,
        cache_docker=False,
        docker_secrets=None,
        nextflow_pipeline_params="",
        profile="",
        git_creds=None,
        brief=False,
        destination=None,
        extra_args=None
):
    """
    :param repository: URL to a Git repository
    :type repository: string
    :param tag: tag of a given Git repository. If it is not provided, the default branch is used.
    :type tag: string
    :param cache_docker: Pull a remote docker image and store it on the platform
    :type cache_docker: bool
    :param docker_secrets: Dx file id with the private docker registry credentials
    :type docker_secrets: string
    :param nextflow_pipeline_params: Custom Nextflow pipeline parameters
    :type nextflow_pipeline_params: string
    :param profile: Custom Nextflow profile, for more information visit https://www.nextflow.io/docs/latest/config.html#config-profiles
    :type profile: string
    :param brief: Level of verbosity
    :type brief: boolean
    :returns: ID of the created applet

    Runs the Nextflow Pipeline Importer app, which creates a Nextflow applet from a given Git repository.
    """

    def parse_extra_args(args):
        """
        :param args: extra args from command input
        :returns: overridable fields from extra_args
        """
        return {
            target_key: val
            for arg_path, target_key in get_allowed_extra_fields_mapping()
            if (val := get_nested(args, arg_path)) is not None
        }

    extra_args = extra_args or {}
    build_project_id = dxpy.WORKSPACE_ID
    build_folder = None
    input_hash = parse_extra_args(extra_args)
    input_hash["repository_url"] = repository

    # { NPI_input_name: (raw value, transformation function),...}
    input_updates = {
        "repository_tag": (tag, None),
        "config_profile": (profile, None),
        "cache_docker": (cache_docker, None),
        "nextflow_pipeline_params": (nextflow_pipeline_params, None),
        "docker_secrets": (docker_secrets, partial(parse_obj, klass="file")),
        "github_credentials": (git_creds, partial(parse_obj, klass="file")),
    }
    for key, (raw_value, transform) in input_updates.items():
        if raw_value:
            input_hash[key] = transform(raw_value) if transform else raw_value

    if destination:
        build_project_id, build_folder, _ = try_call(resolve_existing_path, destination, expected='folder')
    if build_project_id is None:
        parser.error(
            "Can't create an applet without specifying a destination project; please use the -d/--destination flag to explicitly specify a project")
    nf_builder_job = dxpy.DXApp(name=get_importer_name()).run(app_input=input_hash, project=build_project_id,
                                                              folder=build_folder,
                                                              name="Nextflow build of %s" % (repository), detach=True)

    if not brief:
        print("Started builder job %s" % (nf_builder_job.get_id(),))
    nf_builder_job.wait_on_done(interval=1)
    applet_id, _ = dxpy.get_dxlink_ids(nf_builder_job.describe(fields={"output": True})['output']['output_applet'])
    if not brief:
        print("Created Nextflow pipeline %s" % (applet_id))
    else:
        print(json.dumps(dict(id=applet_id)))
    return applet_id


def prepare_nextflow(
        resources_dir,
        profile,
        region,
        cache_docker=False,
        nextflow_pipeline_params=""
):
    """
    :param resources_dir: Directory with all resources needed for the Nextflow pipeline. Usually directory with user's Nextflow files.
    :type resources_dir: str or Path
    :param profile: Custom Nextflow profile. More profiles can be provided by using comma separated string (without whitespaces).
    :type profile: str
    :param region: The region in which the applet will be built.
    :type region: str
    :param cache_docker: Perform pipeline analysis and cache the detected docker images on the platform
    :type cache_docker: boolean
    :param nextflow_pipeline_params: Custom Nextflow pipeline parameters
    :type nextflow_pipeline_params: string
    :returns: Path to the created dxapp_dir
    :rtype: Path

    Creates files necessary for creating an applet on the Platform, such as dxapp.json and a source file. These files are created in '.dx.nextflow' directory.
    """
    assert os.path.exists(resources_dir)
    if not glob(os.path.join(resources_dir, "*.nf")):
        raise dxpy.app_builder.AppBuilderException(
            "Directory %s does not contain Nextflow file (*.nf): not a valid Nextflow directory" % resources_dir)
    dxapp_dir = tempfile.mkdtemp(prefix=".dx.nextflow")

    custom_inputs = prepare_custom_inputs(schema_file=os.path.join(resources_dir, "nextflow_schema.json"))
    dxapp_content = get_nextflow_dxapp(
        custom_inputs=custom_inputs,
        resources_dir=resources_dir,
        region=region,
        profile=profile,
        cache_docker=cache_docker,
        nextflow_pipeline_params=nextflow_pipeline_params
    )
    exec_content = get_nextflow_src(custom_inputs=custom_inputs, profile=profile, resources_dir=resources_dir)
    shutil.copytree(get_template_dir(), dxapp_dir, dirs_exist_ok=True)
    write_dxapp(dxapp_dir, dxapp_content)
    write_exec(dxapp_dir, exec_content)
    create_readme(resources_dir, dxapp_dir)
    return dxapp_dir


def prepare_custom_inputs(schema_file="./nextflow_schema.json"):
    """
    :param schema_file: path to nextflow_schema.json file
    :type schema_file: str or Path
    :returns: list of custom inputs defined with DNAnexus datatype 
    :rtype: list
    Creates custom input list from nextflow_schema.json that
    will be added in dxapp.json inputSpec field
    """

    def get_dx_type(nf_type, nf_format=None):
        types = {
            "string": "string",
            "integer": "int",
            "number": "float",
            "boolean": "boolean",
            "object": "hash"
        }
        str_types = {
            "file-path": "file",
            "directory-path": "string",  # So far we will stick with strings dx://...
            "path": "string"
        }
        if nf_type == "string" and nf_format in str_types:
            return str_types[nf_format]
        elif nf_type in types:
            return types[nf_type]
        raise Exception("type {} is not supported by DNAnexus".format(nf_type))

    inputs = []
    if not os.path.exists(schema_file):
        return inputs

    with open(schema_file, "r") as fh:
        schema = json.load(fh)
    defs_key = "definitions" if "definitions" in schema else "$defs" if "$defs" in schema else {}
    for d_key, d_schema in schema.get(defs_key).items():
        required_inputs = d_schema.get("required", [])
        for property_key, property in d_schema.get("properties", {}).items():
            dx_input = {}
            dx_input["name"] = property_key
            dx_input["title"] = dx_input['name']
            if "default" in property:
                dx_input["help"] = "Default value:{}\n".format(property.get("default", ""))
            if "help_text" in property:
                dx_input["help"] = dx_input.get("help", "") + property.get('help_text', "")
            dx_input["hidden"] = property.get('hidden', False)
            dx_input["class"] = get_dx_type(property.get("type"), property.get("format"))
            dx_input["optional"] = True
            if property_key not in required_inputs:
                dx_input["help"] = "(Nextflow pipeline optional) {}".format(dx_input.get("help", ""))
                inputs.append(dx_input)
            else:
                dx_input["help"] = "(Nextflow pipeline required) {}".format(dx_input.get("help", ""))
                inputs.insert(0, dx_input)

    return inputs



#!/usr/bin/env python3

from __future__ import print_function, unicode_literals, division, absolute_import

# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

# Run manually with python2 and python3 src/python/test/test_create_cohort.py


import json
import unittest
import os
import re
import subprocess
import dxpy
import sys
import hashlib
import uuid
from parameterized import parameterized
from dxpy.bindings import DXRecord, DXProject

from dxpy.cli.dataset_utilities import (
    resolve_validate_dx_path,
    validate_project_access,
    resolve_validate_record_path,
    raw_cohort_query_api_call
)
from dxpy.dx_extract_utils.cohort_filter_payload import (
    generate_pheno_filter,
    cohort_filter_payload,
    cohort_final_payload,
)

dirname = os.path.dirname(__file__)
payloads_dir = os.path.join(dirname, "create_cohort_test_files/payloads/")

python_version = sys.version_info.major

class DescribeDetails:
    """
    Strictly parses describe output into objects attributes.
    ID                                record-GYvjYf00F69fGVYkgXqfzfQ2
    Class                             record
    ...
    Size                              620
    """
    def __init__(self, describe):
        self.parse_atributes(describe)

    def parse_atributes(self, describe):
        for line in describe.split("\n"):
            if line != "":
                p_line = line.split("   ")
                setattr(self, p_line[0].replace(" ", "_"), p_line[-1].strip(" "))


class TestCreateCohort(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        proj_name = "dx-toolkit_test_data"
        proj_id = list(
            dxpy.find_projects(describe=False, level="VIEW", name=proj_name)
        )[0]["id"]
        cls.general_input_dir = os.path.join(dirname, "create_cohort_test_files/input/")
        # cls.general_output_dir = os.path.join(dirname, "create_cohort_test_files/output/")
        cls.payloads_dir = payloads_dir

        # TODO: setup project folders
        cls.proj_id = proj_id
        cls.temp_proj = DXProject()
        cls.temp_proj.new(name="temp_test_create_cohort_{}".format(uuid.uuid4()))
        cls.temp_proj_id = cls.temp_proj._dxid
        dxpy.config["DX_PROJECT_CONTEXT_ID"] = cls.temp_proj_id
        cls.test_record_geno = "{}:/Create_Cohort/create_cohort_geno_dataset".format(proj_name)
        cls.test_record_pheno = "{}:/Create_Cohort/create_cohort_pheno_dataset".format(proj_name)
        cls.test_invalid_rec_type = "{}:/Create_Cohort/non_dataset_record".format(proj_name)
        with open(
            os.path.join(dirname, "create_cohort_test_files", "usage_message.txt"), "r"
        ) as infile:
            cls.usage_message = infile.read()

        cls.maxDiff = None

    @classmethod
    def tearDownClass(cls):
        print("Remmoving temporary testing project {}".format(cls.temp_proj_id))
        cls.temp_proj.destroy()
        del cls.temp_proj

    def find_record_id(self, text): 
        match = re.search(r"\b(record-[A-Za-z0-9]{24})\b", text)
        if match:
            return match.group(0)
        
    def is_record_id(self, text):
        return bool(re.match(r"^(record-[A-Za-z0-9]{24})",text))
    
    def build_command(self, path=None, input_record=None, cohort_ids=None, cohort_ids_file=None):
        command = [
            "dx",
            "create_cohort",
            "--from",
            input_record
        ]
        if path:
            command.append(path)

        if cohort_ids:
            command.extend(["--cohort-ids", cohort_ids])
        
        if cohort_ids_file:
            command.extend(["--cohort-ids-file", cohort_ids_file])
        
        return command
    
    # Test the message printed on stdout when the --help flag is provided
    # This message is also printed on every error caught by argparse, before the specific message
    def test_help_text(self):
        expected_result = self.usage_message
        command = "dx create_cohort --help"

        process = subprocess.check_output(command, shell=True) 

        self.assertEqual(expected_result, process.decode())

    # testing that command accepts file with sample ids
    def test_accept_file_ids(self):
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = self.test_record_pheno,
            cohort_ids_file = "{}sample_ids_valid_pheno.txt".format(self.general_input_dir)
        )

        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        stdout, stderr = process.communicate()

        # testing if record object was created, retrieve record_id from stdout
        record_id = self.find_record_id(stdout)
        self.assertTrue(bool(record_id), "Record object was not created")
        

    # EM-1
    # testing resolution of invalid sample_id provided via file
    def test_accept_file_ids_negative(self):
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = self.test_record_pheno,
            cohort_ids_file = "{}sample_ids_wrong.txt".format(self.general_input_dir)
        )
        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        stderr = process.communicate()[1]
        expected_error = (
            "The following supplied IDs do not match IDs in the main entity of dataset"
        )
        self.assertTrue(expected_error in stderr, msg = stderr)

    def test_accept_cli_ids(self):
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = self.test_record_geno,
            cohort_ids = " sample_1_1 , sample_1_10 "
        )
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        stdout, stderr = process.communicate()

        # testing if record object was created, retrieve record_id from stdout
        record_id = self.find_record_id(stdout)
        self.assertTrue(bool(record_id), "Record object was not created")


    # EM-1
    # Supplied IDs do not match IDs of main entity in Dataset/Cohort
    def test_accept_cli_ids_negative(self):
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = self.test_record_geno,
            cohort_ids = "wrong,sample,id"
        )
        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        stderr = process.communicate()[1]
        expected_error = (
            "The following supplied IDs do not match IDs in the main entity of dataset"
        )
        self.assertTrue(expected_error in stderr, msg = stderr)


    # EM-2
    # The structure of '--from' is invalid. This should be able to be reused from other dx functions
    def test_errmsg_invalid_path(self):
        bad_record = "record-badrecord"
        expected_error_message = (
            'Unable to resolve "{}" to a data object or folder name in'.format(bad_record)
        )
        command = self.build_command(
            input_record = "{}:{}".format(self.proj_id, bad_record),
            cohort_ids = "id1,id2"
        )

        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True
        )

        err_msg = process.communicate()[1]

        # stdout should be the first element in this list and stderr the second
        self.assertIn(expected_error_message, err_msg.strip("\n"))

    # EM-3
    # The user does not have access to the object
    def test_errmsg_no_data_access(self):
        pass

    # EM-4
    # The record id or path is not a cohort or dataset
    # This should fail before the id validity check
    def test_errmsg_not_cohort_dataset(self):

        expected_error_message = "{}: Invalid path. The path must point to a record type of cohort or dataset".format(
            self.test_invalid_rec_type
        )
        command = self.build_command(
            input_record = self.test_invalid_rec_type,
            cohort_ids = "fakeid"
        )
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE ,stderr=subprocess.PIPE, universal_newlines=True
        )

        # stdout should be the first element in this list and stderr the second
        self.assertEqual(expected_error_message, process.communicate()[1].strip())

    # EM-5
    # The record id or path is a cohort or dataset but is invalid (maybe corrupted, descriptor not accessible...etc)
    def test_errmsg_invalid_record(self):
        pass

    # EM-6
    # The record id or path is a cohort or dataset but the version is less than 3.0.
    def test_errmsg_dataset_version(self):
        pass

    # EM-7
    # If PATH is of the format `project-xxxx:folder/` and the project does not exist
    def test_errmsg_project_not_exist(self):
        bad_project = "project-notarealproject7843k2Jq"
        expected_error_message = 'ResolutionError: Could not find a project named "{}"'.format(
            bad_project
        )
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = "{}:/".format(bad_project),
            cohort_ids = "id_1,id_2"
        )
        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        err_msg = process.communicate()[1]
        self.assertIn(expected_error_message, err_msg.strip("\n"))

    # EM-8
    # If PATH is of the format `project-xxxx:folder/` and the user does not have CONTRIBUTE or ADMINISTER access
    # Note that this is the PATH that the output cohort is being created in, not the input dataset or cohort
    def test_errmsg_no_path_access(self):
        pass

    # EM-9
    # If PATH is of the format `folder/subfolder/` and the path does not exist
    def test_errmsg_subfolder_not_exist(self):
        bad_path = "{}:Create_Cohort/missing_folder/file_name".format(self.proj_id)
        expected_error_message = "The folder: {} could not be found in the project: {}".format(
            "/Create_Cohort/missing_folder", self.proj_id
        )
        command = self.build_command(
            path = bad_path,
            input_record = self.test_record_pheno,
            cohort_ids = "patient_1,patient_2"
        )
        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        # split("\n")[-1] is added to ignore the warning message that gets added
        # since DX_PROJECT_CONTEXT_ID environment variable is manually updated in setup class
        err_msg = process.communicate()[1].strip("\n").split("\n")[-1]
        self.assertEqual(expected_error_message, err_msg)

    # EM-10
    # If both --cohort-ids and --cohort-ids-file are supplied in the same call
    # The file needs to exist for this check to be performed
    def test_errmsg_incompat_args(self):
        expected_error_message = "dx create_cohort: error: argument --cohort-ids-file: not allowed with argument --cohort-ids"
        command = self.build_command(
            input_record = self.test_record_pheno,
            cohort_ids = "id1,id2",
            cohort_ids_file = os.path.join(self.general_input_dir, "sample_ids_10.txt")
        )
        process = subprocess.Popen(
            command, stderr=subprocess.PIPE, universal_newlines=True
        )
        err_msg = process.communicate()[1]
        self.assertIn(expected_error_message, err_msg)

    # EM-11 The vizserver returns an error when attempting to validate cohort IDs
    def test_vizserver_error(self):
        pass

    def test_raw_cohort_query_api(self):
        test_payload = {
            "filters": {
                "pheno_filters": {
                    "compound": [
                        {
                            "name": "phenotype",
                            "logic": "and",
                            "filters": {
                                "patient$patient_id": [
                                    {
                                        "condition": "in",
                                        "values": [
                                            "patient_1",
                                            "patient_2",
                                            "patient_3"
                                        ]
                                    }
                                ]
                            }
                        }
                    ],
                    "logic": "and"
                },
                "logic": "and"
            },
            "project_context": self.proj_id
        }

        expected_results = "SELECT `patient_1`.`patient_id` AS `patient_id` FROM `database_yyyyyyyyyyyyyyyyyyyyyyyy__create_cohort_pheno_database`.`patient` AS `patient_1` WHERE `patient_1`.`patient_id` IN ('patient_1', 'patient_2', 'patient_3');"

        from_project, entity_result, resp, dataset_project = resolve_validate_record_path(self.test_record_pheno)
        sql = raw_cohort_query_api_call(resp, test_payload)
        self.assertEqual(expected_results, re.sub(r"\bdatabase_\w{24}__\w+", "database_yyyyyyyyyyyyyyyyyyyyyyyy__create_cohort_pheno_database", sql))

    def test_create_pheno_filter(self):
        """Verifying the correctness of created filters by examining this flow:
            1. creating the filter with: dxpy.dx_extract_utils.cohort_filter_payload.generate_pheno_filter
            2. obtaining sql with: dxpy.cli.dataset_utilities.raw_cohort_query_api_call
            3. creating record with obtained sql and the filter by: dxpy.bindings.dxrecord.new_dxrecord
        """

        # test creating pheno filter
        values = ["patient_1", "patient_2", "patient_3"]
        entity = "patient"
        field = "patient_id"
        filters = {
            "pheno_filters": {
                "compound": [
                    {
                        "name": "phenotype",
                        "logic": "and",
                        "filters": {
                            "patient$patient_id": [
                                {
                                    "condition": "in",
                                    "values": ["patient_1", "patient_2", "patient_6"],
                                }
                            ]
                        },
                    }
                ],
                "logic": "and",
            },
            "logic": "and",
        }
        expected_filter = {
            "pheno_filters": {
                "compound": [
                    {
                        "name": "phenotype",
                        "logic": "and",
                        "filters": {
                            "patient$patient_id": [
                                {
                                    "condition": "in",
                                    "values": ["patient_1", "patient_2"]
                                }
                            ]
                        },
                    }
                ],
                "logic": "and",
            },
            "logic": "and",
        }
        expected_sql = "SELECT `patient_1`.`patient_id` AS `patient_id` FROM `database_yyyyyyyyyyyyyyyyyyyyyyyy__create_cohort_pheno_database`.`patient` AS `patient_1` WHERE `patient_1`.`patient_id` IN ('patient_1', 'patient_2');"
        lambda_for_list_conv = lambda a, b: a+[str(b)]
        
        generated_filter = generate_pheno_filter(values, entity, field, filters, lambda_for_list_conv)
        self.assertEqual(expected_filter, generated_filter)

        # Testing raw cohort query api
        resp = resolve_validate_record_path(self.test_record_pheno)[2]
        payload = {"filters": generated_filter, "project_context": self.proj_id}

        sql = raw_cohort_query_api_call(resp, payload)
        self.assertEqual(expected_sql, re.sub(r"\bdatabase_\w{24}__\w+", "database_yyyyyyyyyyyyyyyyyyyyyyyy__create_cohort_pheno_database", sql))

        # Testing new record with generated filter and sql
        details = {
            "databases": [resp["databases"]],
            "dataset": {"$dnanexus_link": resp["dataset"]},
            "description": "",
            "filters": generated_filter,
            "schema": "create_cohort_schema",
            "sql": sql,
            "version": "3.0",
        }


        new_record = dxpy.bindings.dxrecord.new_dxrecord(
            details=details,
            project=self.temp_proj_id,
            name=None,
            types=["DatabaseQuery", "CohortBrowser"],
            folder="/",
            close=True,
        )
        new_record_details = new_record.get_details()
        new_record.remove()
        e = None
        self.assertTrue(isinstance(new_record, DXRecord))
        self.assertEqual(new_record_details, details, "Details of created record does not match expected details.")

    @parameterized.expand(
        os.path.splitext(file_name)[0] for file_name in sorted(os.listdir(os.path.join(payloads_dir, "raw-cohort-query_input")))
    )
    def test_cohort_filter_payload(self, payload_name):
        with open(os.path.join(self.payloads_dir, "input_parameters", "{}.json".format(payload_name))) as f:
            input_parameters = json.load(f)
        values = input_parameters["values"]
        entity = input_parameters["entity"]
        field = input_parameters["field"]
        project_context = input_parameters["project"]

        with open(os.path.join(self.payloads_dir, "visualize_response", "{}.json".format(payload_name))) as f:
            visualize_response = json.load(f)
        filters = visualize_response.get("filters", {})
        base_sql = visualize_response.get("baseSql", visualize_response.get("base_sql"))
        lambda_for_list_conv = lambda a, b: a+[str(b)]

        test_payload = cohort_filter_payload(values, entity, field, filters, project_context, lambda_for_list_conv, base_sql)

        with open(os.path.join(self.payloads_dir, "raw-cohort-query_input", "{}.json".format(payload_name))) as f:
            valid_payload = json.load(f)

        self.assertDictEqual(test_payload, valid_payload)

    @parameterized.expand(
        os.path.splitext(file_name)[0] for file_name in sorted(os.listdir(os.path.join(payloads_dir, "dx_new_input")))
    )
    def test_cohort_final_payload(self, payload_name):
        name = None

        with open(os.path.join(self.payloads_dir, "input_parameters", "{}.json".format(payload_name))) as f:
            input_parameters = json.load(f)
        folder = input_parameters["folder"]
        project = input_parameters["project"]

        with open(os.path.join(self.payloads_dir, "visualize_response", "{}.json".format(payload_name))) as f:
            visualize = json.load(f)
        dataset = visualize["dataset"]
        databases = visualize["databases"]
        schema = visualize["schema"]
        base_sql = visualize.get("baseSql", visualize.get("base_sql"))
        combined = visualize.get("combined")

        with open(os.path.join(self.payloads_dir, "raw-cohort-query_input", "{}.json".format(payload_name))) as f:
            filters = json.load(f)["filters"]

        with open(os.path.join(self.payloads_dir, "raw-cohort-query_output", "{}.sql".format(payload_name))) as f:
            sql = f.read()

        test_output = cohort_final_payload(name, folder, project, databases, dataset, schema, filters, sql, base_sql, combined)

        with open(os.path.join(self.payloads_dir, "dx_new_input", "{}.json".format(payload_name))) as f:
            valid_output = json.load(f)

        valid_output["name"] = None

        self.assertDictEqual(test_output, valid_output)

    def test_brief_verbose(self):
        command = self.build_command(
            path = "{}:/".format(self.temp_proj_id),
            input_record = self.test_record_geno,
            cohort_ids = "sample_1_1,sample_1_10"
        )

        for stdout_mode in ["--verbose", "--brief", ""]:
            cmd = command + [stdout_mode] if stdout_mode != "" else command
            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            stdout, stderr = process.communicate()
            if stdout_mode == "--brief":
                record_id = stdout.strip("\n").strip(" ")
                self.assertTrue(self.is_record_id(record_id), 
                                "Brief stdout has to be a record-id"
                                )
            elif stdout_mode == "--verbose":
                self.assertIn(
                    "Details", stdout, "Verbose stdout has to contain 'Details' string"
                )
            else:
                self.assertIn(
                    "Types", stdout, "Default stdout has to contain 'Types' string"
                )


    def test_path_upload_access(self):
        # Having at least UPLOAD access to a project
        err_msg = validate_project_access(self.temp_proj_id)
        self.assertIsNone(err_msg)

    def test_path_upload_access_negative(self):
        #TODO: delegate this to QE
        pass

    def test_path_options(self):
        """
        Testing different path formats.
        Various path options and expected results are parametrized. 
        The dictionary `expected_in_out_pairs` expects form: {"<path>": (<results tuple>)}
        """
        self.temp_proj.new_folder("/folder/subfolder", parents=True)
        
        expected_in_out_pairs = {
            "{}:/".format(self.proj_id): (self.proj_id, "/", None, None),
            "{}:/folder/subfolder/record_name1".format(self.temp_proj_id): (self.temp_proj_id, "/folder/subfolder", "record_name1", None ),
            "record_name": (self.temp_proj_id, "/", "record_name", None),
            "/folder/record_name": (self.temp_proj_id, "/folder", "record_name", None),
            "/folder/subfolder/record_name": (
                self.temp_proj_id,
                "/folder/subfolder",
                "record_name",
                None,
            ),
            "/folder/subfolder/no_exist/record_name": (
                self.temp_proj_id,
                "/folder/subfolder/no_exist",
                "record_name",
                "The folder: /folder/subfolder/no_exist could not be found in the project: {}".format(
                    self.temp_proj_id
                ),
            ),
            "/folder/": (self.temp_proj_id, "/folder", None, None),
        }

        for path, expected_result in expected_in_out_pairs.items():
            result = resolve_validate_dx_path(path)
            self.assertEqual(result, expected_result) 

    def test_path_options_negative(self):
        expected_result = (
            self.temp_proj_id,
            "/folder/subfolder/no_exist",
            "record_name",
            "The folder: /folder/subfolder/no_exist could not be found in the project: {}".format(
                self.temp_proj_id
            )
        )
        result = resolve_validate_dx_path("/folder/subfolder/no_exist/record_name") 
        self.assertEqual(result, expected_result)

    def test_path_options_cli(self):
        """
        Testing different path formats. 
        Focusing on default values with record name or folder not specified. 
        """
        # create subfolder structure
        self.temp_proj.new_folder("/folder/subfolder", parents=True)
        # set cwd
        dxpy.config['DX_CLI_WD'] = "/folder"

        command = self.build_command(
            input_record = self.test_record_geno,
            cohort_ids = "sample_1_1,sample_1_10"
        )
        
        path_options = [        
            "record_name2", #Should create record in CWD
            "/folder/subfolder/", #Name of record should be record-id
            "", #Combination of above
        ]
        for path_format in path_options:
            cmd = command[:]
            if path_format != "":
                cmd.insert(2, path_format)

            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )
            stdout = process.communicate()[0]
            desc = DescribeDetails(stdout)

            if path_format == "record_name2":
                self.assertEqual(desc.Folder, "/folder")
                self.assertEqual(desc.Project, self.temp_proj_id)
                self.assertEqual(desc.Name, "record_name2")
            elif path_format =="/folder/subfolder/":
                self.assertEqual(desc.Folder, "/folder/subfolder")
                self.assertEqual(desc.Project, self.temp_proj_id)
                self.assertTrue(self.is_record_id(desc.Name), 
                                "Record name should be a record-id"
                                )
            elif path_format =="":
                self.assertEqual(desc.Folder, "/folder")
                self.assertEqual(desc.Project, self.temp_proj_id)
                self.assertTrue(self.is_record_id(desc.Name), 
                                "Record name should be a record-id"
                                )
                
                
    def test_path_options_cli_negative(self):
        command = self.build_command(
            path = "/folder/subfolder/no_exist/record_name",
            input_record = self.test_record_geno,
            cohort_ids = "sample_1_1,sample_1_10"
        )
        process = subprocess.Popen(
                command,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
                universal_newlines=True,
            )
        stderr = process.communicate()[1]
        self.assertEqual(
            "The folder: /folder/subfolder/no_exist could not be found in the project: {}".format(
                self.temp_proj_id
                ), 
            stderr.strip("\n"))


if __name__ == "__main__":
    unittest.main()




# Copyright (C) 2014-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

'''
App, Applet, and Workflow Unbuilder
+++++++++++++

Contains utility methods for "decompiling" an app, applet, or workflow
from the platform and generating an equivalent source directory.

'''

from __future__ import print_function, unicode_literals, division, absolute_import

import collections
import json
import os
import sys
import tarfile
import shutil

import dxpy
from .. import get_handler, download_dxfile
from ..exceptions import err_exit, DXError
from .pretty_print import flatten_json_array


def _recursive_cleanup(foo):
    """
    Aggressively cleans up things that look empty.
    """
    if isinstance(foo, dict):
        for (key, val) in list(foo.items()):
            if isinstance(val, dict):
                _recursive_cleanup(val)
            if val == "" or val == [] or val == {}:
                del foo[key]


def _write_json_file(filename, json_content):
    with open(filename, "w") as f:
        f.write(flatten_json_array(json.dumps(json_content, indent=2, separators=(',', ': ')), "patterns"))
        f.write('\n')


def _write_simple_file(filename, content):
    with open(filename, "w") as f:
        f.write(content)


def _dump_workflow(workflow_obj, describe_output={}):
    dxworkflow_json_keys = ['name', 'title', 'summary', 'dxapi', 'version',
                            'outputFolder', 'treeTurnaroundTimeThreshold']
    dxworkflow_json_stage_keys = ['id', 'name', 'executable', 'folder', 'input',
                                  'executionPolicy', 'systemRequirements']

    dxworkflow_json = collections.OrderedDict()
    for key in dxworkflow_json_keys:
        if key in describe_output and describe_output[key]:
            dxworkflow_json[key] = describe_output[key]

    for key in ('inputs', 'outputs'):
        if key in describe_output and describe_output[key] is not None:
            dxworkflow_json[key] = describe_output[key]
    stages = describe_output.get("stages", ())
    new_stages = []
    for stage in stages:
        new_stage = collections.OrderedDict()
        for key in dxworkflow_json_stage_keys:
            if key in stage and stage[key]:
                new_stage[key] = stage[key]
        new_stages.append(new_stage)
    dxworkflow_json["stages"] = new_stages

    # Create dxworkflow.json, Readme.md files
    _write_json_file("dxworkflow.json", dxworkflow_json)
    readme = describe_output.get("description", "")
    if readme:
        _write_simple_file("Readme.md", readme)


def _dump_app_or_applet(executable, omit_resources=False, describe_output={}):
    info = executable.get()

    if info["runSpec"]["interpreter"] == "bash":
        suffix = "sh"
    elif info["runSpec"]["interpreter"].startswith("python"):
        suffix = "py"
    else:
        print("Sorry, I don\'t know how to get executables with interpreter {}.\n".format(
            info["runSpec"]["interpreter"]), file=sys.stderr)
        sys.exit(1)

    # Entry point script
    script = "src/code.{}".format(suffix)
    os.mkdir("src")
    with open(script, "w") as f:
        f.write(info["runSpec"]["code"])

    def make_cluster_bootstrap_script_file(region, entry_point, code, suffix):
        """
        Writes the string `code` into a file at the relative path
        "src/<region>_<entry_point>_clusterBootstrap.<suffix>"
        """
        script_name = "src/{}_{}_clusterBootstrap.{}".format(region, entry_point, suffix)
        with open(script_name, "w") as f:
            f.write(code)
        return script_name
    
    # Get regions where the user's billTo are permitted
    try:
        bill_to = dxpy.api.user_describe(dxpy.whoami())['billTo']
        permitted_regions = set(dxpy.DXHTTPRequest('/' + bill_to + '/describe', {}).get("permittedRegions"))
    except DXError:
        print("Failed to get permitted regions where {} can perform billable activities.\n".format(bill_to), file=sys.stderr)
        sys.exit(1)    
    
    # when applet/app is built, the runSpec is initialized with fields "interpreter" and "bundledDependsByRegion"
    # even when we don't have any bundledDepends in dxapp.json
    enabled_regions = set(info["runSpec"]["bundledDependsByRegion"].keys())
    if not enabled_regions.issubset(permitted_regions):
        print("Region(s) {} are not among the permitted regions of {}. Resources from these regions will not be available.".format(
            ", ".join(enabled_regions.difference(permitted_regions)), bill_to), file=sys.stderr )
    # Update enabled regions
    enabled_regions.intersection_update(permitted_regions)
    
    # Start downloading when not omitting resources
    deps_downloaded = set()
    if not omit_resources:
        download_completed = False
        
        # Check if at least one region is enabled 
        if not enabled_regions:
            raise DXError(
                "Cannot download resources of the requested executable {} since it is not available in any of the billable regions. "
                "You can use the --omit-resources flag to skip downloading the resources. ".format(info["name"]))
        
        # Pick a source region. The current selected region is preferred
        try:
            current_region = dxpy.api.project_describe(dxpy.WORKSPACE_ID, input_params={"fields": {"region": True}})["region"]
        except:
            current_region = None

        if current_region in enabled_regions:
            source_region  = current_region
            print("Trying to download resources from the current region {}...".format(source_region), file=sys.stderr)
        else:
            source_region = list(enabled_regions)[0]
            print("Trying to download resources from one of the enabled region {}...".format(source_region), file=sys.stderr)

        # When an app(let) is built the following dependencies are added as bundledDepends:
        # 1. bundledDepends explicitly specified in the dxapp.json
        # 2. resources (contents of resources directory added as bundledDepends)
        # 3. assetDepends in the dxapp.json (with their record IDs translated into file IDs)
        #
        # To get the resources, we will tranverse the bundleDepends list in the source region and do the following:
        # - If an file ID refers to an AssetBundle tarball (a file with the "AssetBundle" property), 
        #   skip downloading the file, and keep this file ID as a bundledDepends in the final dxapp.json
        # - Otherwise, download the file and remove this ID from the bundledDepends list in the final dxapp.json

        def untar_strip_leading_slash(tarfname, path):
            with tarfile.open(tarfname) as t:
                for m in t.getmembers():
                    if m.name.startswith("/"):
                        m.name = m.name[1:]
                    t.extract(m, path)
            t.close()

        created_resources_directory = False
        # Download resources from the source region      
        for dep in info["runSpec"]["bundledDependsByRegion"][source_region]:
            try: 
                file_handle = get_handler(dep["id"])
                handler_id = file_handle.get_id()
                # if dep is not a file (record etc.), check the next dep
                if not isinstance(file_handle, dxpy.DXFile):
                    continue
                
                # check if the file is an asset dependency
                # if so, skip downloading
                if file_handle.get_properties().get("AssetBundle"):
                    continue

                # if the file is a bundled dependency, try downloading it
                if not created_resources_directory:
                    os.mkdir("resources")
                    created_resources_directory = True
                
                fname = "resources/{}.tar.gz" .format(handler_id)
                download_dxfile(handler_id, fname)
                print("Unpacking resource {}".format(dep.get("name")), file=sys.stderr)

                untar_strip_leading_slash(fname, "resources")
                os.unlink(fname)
                # add dep name to deps_downloaded set
                deps_downloaded.add(dep.get("name"))
                
            except DXError:
                print("Failed to download {} from region {}.".format(handler_id, source_region),
                        file=sys.stderr)
                # clean up deps already downloaded and quit downloading
                deps_downloaded.clear()
                shutil.rmtree("resources")
                break
        # if all deps have been checked without an error, mark downloading as completed
        else: # for loop finished with no break
            download_completed = True                    
    
        # Check if downloading is completed in one of the enabled regions
        # if so, files in deps_downloaded will not shown in dxapp.json
        # if not, deps_downloaded is an empty set. So ID of all deps will be in dxapp.json
        if not download_completed:
            print("Downloading resources from region {} failed. "
                "Please try downloading with their IDs in dxapp.json, "
                "or skip downloading resources entirely by using the --omit-resources flag.".format(source_region), file=sys.stderr)

    # TODO: if output directory is not the same as executable name we
    # should print a warning and/or offer to rewrite the "name"
    # field in the 'dxapp.json'
    dxapp_json = collections.OrderedDict()
    all_keys = executable._get_required_keys() + executable._get_optional_keys()
    for key in all_keys:
        if key in executable._get_describe_output_keys() and key in describe_output:
            dxapp_json[key] = describe_output[key]
        if key in info:
            dxapp_json[key] = info[key]
    if info.get("hidden", False):
        dxapp_json["hidden"] = True

    # TODO: inputSpec and outputSpec elements should have their keys
    # printed in a sensible (or at least consistent) order too

    # Un-inline code
    del dxapp_json["runSpec"]["code"]
    dxapp_json["runSpec"]["file"] = script

    # Ordering input/output spec keys
    ordered_spec_keys = ("name", "label", "help", "class", "type", "patterns", "optional", "default", "choices",
                         "suggestions", "group")
    for spec_key in "inputSpec", "outputSpec":
        if spec_key not in dxapp_json.keys():
            continue
        for i, spec in enumerate(dxapp_json[spec_key]):
            ordered_spec = collections.OrderedDict()
            # Adding keys, for which the ordering is defined
            for key in ordered_spec_keys:
                if key in spec.keys():
                    ordered_spec[key] = spec[key]
            # Adding the rest of the keys
            for key in spec.keys():
                if key not in ordered_spec_keys:
                    ordered_spec[key] = spec[key]
            dxapp_json[spec_key][i] = ordered_spec

    # Remove dx-toolkit from execDepends
    dx_toolkit = {"name": "dx-toolkit", "package_manager": "apt"}
    if dx_toolkit in dxapp_json["runSpec"].get("execDepends", ()):
        dxapp_json["runSpec"]["execDepends"].remove(dx_toolkit)

    # "dx build" parses the "regionalOptions" key from dxapp.json into the
    # "runSpec.systemRequirements" field of applet/new.
    # "dx get" should parse the "systemRequirementsByRegion" field from
    # the response of /app-x/get or /applet-x/get into the "regionalOptions"
    # key in dxapp.json.
    dxapp_json["regionalOptions"] = {}
    for region in enabled_regions:
        dxapp_json["regionalOptions"][region] = {}
        if "systemRequirementsByRegion" in dxapp_json['runSpec']:
            region_sys_reqs = dxapp_json['runSpec']['systemRequirementsByRegion'][region]

            # handle cluster bootstrap scripts if any are present
            for entry_point in region_sys_reqs:
                try:
                    bootstrap_script = region_sys_reqs[entry_point]['clusterSpec']['bootstrapScript']
                    filename = make_cluster_bootstrap_script_file(region,
                                                                  entry_point,
                                                                  bootstrap_script,
                                                                  suffix)
                    region_sys_reqs[entry_point]['clusterSpec']['bootstrapScript'] = filename
                except KeyError:
                    # either no "clusterSpec" or no "bootstrapScript" within "clusterSpec"
                    continue

            dxapp_json["regionalOptions"][region]["systemRequirements"]=region_sys_reqs

        region_depends = dxapp_json["runSpec"]["bundledDependsByRegion"][region]
        region_bundle_depends = [d for d in region_depends if d["name"] not in deps_downloaded]
        if region_bundle_depends:
            dxapp_json["regionalOptions"][region]["bundledDepends"]=region_bundle_depends
            
    # Remove "bundledDependsByRegion" and "bundledDepends" field from "runSpec".
    # assetDepends and bundledDepends data are stored in regionalOptions instead.
    dxapp_json["runSpec"].pop("bundledDependsByRegion", None)
    dxapp_json["runSpec"].pop("bundledDepends", None)
    # systemRequirementsByRegion data is stored in regionalOptions,
    # systemRequirements is ignored
    dxapp_json["runSpec"].pop("systemRequirementsByRegion",None)
    dxapp_json["runSpec"].pop("systemRequirements",None)

    # Cleanup of empty elements. Be careful not to let this step
    # introduce any semantic changes to the app specification. For
    # example, an empty input (output) spec is not equivalent to a
    # missing input (output) spec.
    if 'runSpec' in dxapp_json:
        _recursive_cleanup(dxapp_json['runSpec'])
    if 'access' in dxapp_json:
        _recursive_cleanup(dxapp_json['access'])
    for key in executable._get_cleanup_keys():
        if key in dxapp_json and not dxapp_json[key]:
            del dxapp_json[key]

    readme = info.get("description", "")
    devnotes = info.get("developerNotes", "")

    # Write dxapp.json, Readme.md, and Readme.developer.md
    _write_json_file("dxapp.json", dxapp_json)
    if readme:
        _write_simple_file("Readme.md", readme)
    if devnotes:
        _write_simple_file("Readme.developer.md", devnotes)

def dump_executable(executable, destination_directory, omit_resources=False, describe_output={}):
    """
    Reconstitutes an app, applet, or a workflow into a directory that would
    create a functionally identical executable if "dx build" were run on it.
    destination_directory will be the root source directory for the
    executable.

    :param executable: executable, i.e. app, applet, or workflow, to be dumped
    :type executable: DXExecutable (either of: DXApp, DXApplet, DXWorkflow, DXGlobalWorkflow)
    :param destination_directory: an existing, empty, and writable directory
    :type destination_directory: str
    :param omit_resources: if True, executable's resources will not be downloaded
    :type omit_resources: boolean
    :param describe_output: output of a describe API call on the executable
    :type describe_output: dictionary
    """
    try:
        old_cwd = os.getcwd()
        os.chdir(destination_directory)
        if isinstance(executable, dxpy.DXWorkflow):
            _dump_workflow(executable, describe_output)
        elif isinstance(executable, dxpy.DXGlobalWorkflow):
            # Add inputs, outputs, stages. These fields contain region-specific values
            # e.g. files or applets, that's why:
            # * if the workflow is global, we will unpack the underlying workflow
            #   from the region of the current project context
            # * if this is a regular, project-based workflow, we will just use
            #   its description (the describe_output that we already have)
            # Underlying workflows are workflows stored in resource containers
            # of the global workflow (one per each region the global workflow is
            # enabled in). #TODO: add a link to documentation.
            current_project = dxpy.WORKSPACE_ID
            if not current_project:
                raise DXError(
                    'A project needs to be selected to "dx get" a global workflow. You can use "dx select" to select a project')
            region = dxpy.api.project_describe(current_project,
                                               input_params={"fields": {"region": True}})["region"]
            describe_output = executable.append_underlying_workflow_desc(describe_output, region)
            _dump_workflow(executable, describe_output)
        else:
            _dump_app_or_applet(executable, omit_resources, describe_output)
    except:
        err_exit()
    finally:
        os.chdir(old_cwd)
